import os
import os.path as osp
import glob
import pathlib
from pathlib import (
  Path,
  PurePath,
  PurePosixPath)

import logging

from ..validate import (
  ValidationError,
  FileOutsideRootError,
  validating )

from ..norms import (
  norm_path )

from ..path import (
  PathMatcher,
  PathFilter,
  subdir,
  combine_ignore_patterns,
  contains )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def dist_iter(*,
  include,
  ignore,
  root ):

  patterns = PathFilter(
    patterns = ignore )

  for i, incl in enumerate(include):
    src = incl.src
    dst = incl.dst
    _ignore = incl.ignore

    _ignore_patterns = combine_ignore_patterns(
      patterns,
      PathFilter(
        patterns = _ignore,
        start = src ) )

    if incl.glob:

      cwd = Path.cwd()
      try:
        os.chdir(src) #Doesn't seems to be an pathlib way of changing working direcotry
        matches = glob.glob(incl.glob, recursive = True)
      finally:
        os.chdir(cwd)

      for match in matches:
        _src = src.joinpath(match)
        # re-base the dst path, path relative to src == path relative to dst
        _dst = dst.joinpath(match)

        yield ( i, _src, _dst, _ignore_patterns, False )

    else:

      yield ( i, src, dst, _ignore_patterns, True )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def dist_copy(*,
  base_path,
  include,
  ignore,
  dist,
  root = None,
  logger = None ):

  if len(include) == 0:
    return

  logger = logger or logging.getLogger( __name__ )

  with validating(key = 'copy'):

    for i, src, dst, ignore_patterns, individual in dist_iter(
      include = include,
      ignore = ignore,
      root = root ):

      with validating(key = i):

        dst = base_path.joinpath(dst)

        if not individual and ignore_patterns( src.parent, [src.name]):
          logger.debug( f'ignoring: {src}' )
          continue

        src_abs = src.resolve()

        if root and not subdir(root, src_abs, check = False):
          raise FileOutsideRootError(
            f"Must have common path with root:\n  file = \"{src_abs}\"\n  root = \"{root}\"")

        logger.debug(f"dist copy: {src} -> {dst}")

        if src.is_dir():
          dist.copytree(
            src = src,
            dst = dst,
            ignore = ignore_patterns )

        else:
          dist.copyfile(
            src = src,
            dst = dst )



import os
import os.path as osp
import tempfile
import shutil
import pathlib

from pytest import (
  raises )

from partis.pyproj import (
  PathMatcher,
  PathFilter,
  PatternError,
  partition,
  combine_ignore_patterns,
  contains )

pxp = pathlib.PurePosixPath
ntp = pathlib.PureWindowsPath
prp = pathlib.PurePath

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_partition():
  assert partition(lambda x: x > 1, [0, 1, 2]) == ([2], [0, 1])

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_match_escape():
  # These test _match to check the raw string match without normalizing as a path

  # escaped special glob characters
  assert PathMatcher(r'\[]')._match('[]')
  assert PathMatcher(r'\*')._match('*')
  assert PathMatcher(r'\?')._match('?')
  assert PathMatcher(r'\*')._match('*')

  # not escaped
  assert PathMatcher(r'\.')._match(r'\.')
  assert PathMatcher(r'\abc')._match(r'\abc')
  assert PathMatcher(r'.*')._match(r'.*')
  assert PathMatcher(r'.*')._match(r'.*')
  assert PathMatcher(r'.{3}')._match(r'.{3}')

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_match_chr():
  # These test _match to check the raw string match without normalizing as a path
  p = PathMatcher('a?c')
  assert p._match('abc')
  assert p._match('axc')
  assert not p._match('ac')

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_match_chrset():
  # These test _match to check the raw string match without normalizing as a path
  assert PathMatcher('[!]')._match('!')
  assert not PathMatcher('[!!]')._match('!')
  assert not PathMatcher('[^!]')._match('!')
  assert PathMatcher('[]]')._match(']')
  assert not PathMatcher('[!]]')._match(']')
  assert not PathMatcher('[^]]')._match(']')
  assert PathMatcher('[]!]')._match(']')
  assert PathMatcher('[]!]')._match('!')

  assert PathMatcher('[-]')._match('-')
  assert PathMatcher('[--]')._match('-')
  assert PathMatcher('[---]')._match('-')

  assert PathMatcher('[?]')._match('?')
  assert PathMatcher('[*]')._match('*')

  p = PathMatcher('[x-z]')
  assert p._match('x')
  assert p._match('y')
  assert p._match('z')
  assert not p._match('X')
  assert not p._match('w')

  p = PathMatcher('[--0]')
  assert p._match('-')
  assert p._match('.')
  assert not p.posix('/')
  assert p._match('0')

  p = PathMatcher('[b-b]')
  assert p._match('b')
  assert not p._match('a')
  assert not p._match('c')

  # not escaped in character sets
  # bpo-409651
  p = PathMatcher(r'[\]')
  assert p._match('\\')
  assert not p._match('a')

  p = PathMatcher(r'[!\]')
  assert not p._match('\\')
  assert p._match('a')

  with raises(PatternError):
    # must be non-empty
    PathMatcher('[]')

  with raises(PatternError):
    # path separator undefined in char set
    PathMatcher('[/]')

  with raises(PatternError):
    # range is not ordered
    PathMatcher('[z-a]')

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_match_any():
  p = PathMatcher('*.py')
  assert p.posix('.py')
  assert p.posix('a.py')
  assert p.posix('abc.py')
  # * does not match /
  assert not p.posix('a/.py')
  assert not p.posix('a/b/.py')

  # bpo-40480
  p = PathMatcher('*a*a*a*a*a*a*a*a*a*a')
  assert not p.posix('a' * 50 + 'b')

  # pasting multiple segments
  p = PathMatcher('*a*a/*b*b/*c*c')
  assert p.posix('_a_a_a/_b_b_b/_c_c_c')
  assert p.posix('aa/bb/cc')
  assert not p.posix('ab/bc/cd')

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_match():

  p = PathMatcher('a')
  assert str(p) == 'a'
  assert not p.negate
  assert not p.dironly
  assert not p.relative
  assert p.posix('a')

  p = PathMatcher('a/')
  assert not p.negate
  assert p.dironly
  assert not p.relative
  assert p.posix('a')

  p = PathMatcher('/a')
  assert not p.negate
  assert not p.dironly
  assert p.relative
  assert p.posix('a')

  p = PathMatcher('./a')
  assert not p.negate
  assert not p.dironly
  assert p.relative
  assert p.posix('a')

  p = PathMatcher('!a')
  assert p.negate
  assert not p.dironly
  assert not p.relative
  assert p.posix('a')

  p = PathMatcher(r'\!a')
  assert not p.negate
  assert not p.dironly
  assert not p.relative
  assert p.posix('!a')

  p = PathMatcher('a/b')
  assert not p.negate
  assert not p.dironly
  assert p.relative
  assert p.posix('a/b')

  p = PathMatcher('a/b/')
  assert not p.negate
  assert p.dironly
  assert p.relative
  assert p.posix('a/b')

  p = PathMatcher('!a/b')
  assert p.negate
  assert not p.dironly
  assert p.relative
  assert p.posix('a/b')

  p = PathMatcher('!a/')
  assert p.negate
  assert p.dironly
  assert not p.relative
  assert p.posix('a')

  p = PathMatcher('!a/b/')
  assert p.negate
  assert p.dironly
  assert p.relative
  assert p.posix('a/b')

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_match_recurse():

  p = PathMatcher('**/foo')
  assert not p.negate
  assert not p.dironly
  assert p.relative
  assert p.posix('a/b/foo')
  assert p.posix('a/foo')
  assert p.posix('./foo')
  assert p.posix('foo')

  p = PathMatcher('**/foo/bar')
  assert not p.negate
  assert not p.dironly
  assert p.relative
  assert p.posix('a/b/foo/bar')
  assert p.posix('a/foo/bar')
  assert p.posix('foo/bar')

  p = PathMatcher('a/**/b')
  assert not p.negate
  assert not p.dironly
  assert p.relative
  assert p.posix('a/b')
  assert p.posix('a/x/b')
  assert p.posix('a/x/y/b')

  p = PathMatcher('abc/**')
  assert not p.negate
  assert not p.dironly
  assert p.relative
  assert p.posix('abc/a')
  assert p.posix('abc/a/b')

  with raises(PatternError):
    # ** only defined when bounded by /
    # e.g. **/, /**/, or /**
    p = PathMatcher('a**b')

  with raises(PatternError):
    p = PathMatcher('a**')

  with raises(PatternError):
    p = PathMatcher('**b')

  with raises(PatternError):
    p = PathMatcher('a**/b')

  with raises(PatternError):
    p = PathMatcher('**a/b')

  with raises(PatternError):
    p = PathMatcher('a/b**')

  with raises(PatternError):
    p = PathMatcher('a/**b')


#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_filter():

  p = PathFilter()
  assert p.patterns == []
  assert p.start is None
  assert p.filter('.', dnames = ['a'], fnames = ['b']) == set()

  p = PathFilter(['a/', '!b'])
  assert len(p.patterns) == 2
  assert p.patterns[0].posix('a')
  assert p.patterns[1].posix('b')
  assert p.start is None
  assert p.filter('.', dnames = ['a'], fnames = ['b'], feasible = {'b'}) == {'a'}

  p = PathFilter(['x/y'], start = pxp('z'))
  assert len(p.patterns) == 1
  assert p.patterns[0].posix('x/y')
  assert p.start == pxp('z')
  assert p.filter(pxp('z/x'), dnames = [], fnames = ['y']) == {'y'}
  assert p.filter(ntp('z\\x'), dnames = [], fnames = ['y']) == {'y'}

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_file_ignore_patterns():
  ignore_patterns = combine_ignore_patterns(
    PathFilter(['a/', '!b']),
    PathFilter(['x/y'], start = pxp('z')) )

  with tempfile.TemporaryDirectory() as tmpdir:
    a = osp.join(tmpdir,'a')
    x = osp.join(tmpdir,'x')
    y = osp.join(x, 'y')

    os.mkdir(a)
    os.mkdir(x)

    with open( y, 'a'):
      os.utime( y, None )

    assert ignore_patterns('z/x', ['y'])

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
if __name__ == '__main__':
  test_match_any()



#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class PathError(ValueError):
  pass

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def _subdir(_start, _path):

  n = len(_start)

  if len(_path) < n or _path[:n] != _start:
    return None

  return _path[n:]

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def subdir(start, path, check = True):
  """Relative path, restricted to sub-directories.

  Parameters
  ----------
  start : PurePath
    Starting directory.
  path : PurePath
    Directory to compute relative path to, *must* be a sub-directory of `start`.

  Returns
  -------
  rpath : PurePath
    Relative path from `start` to `path`.
  """

  _rpath = _subdir(start.parts, path.parts)

  if _rpath is None:
    if check:
      raise PathError(f"Not a subdirectory of {start}: {path}")

    return None
  
  return type(path)(*_rpath)



import os
import os.path as osp
import tempfile
import shutil
from pathlib import Path

from pytest import (
  raises )

from partis.pyproj.load_module import (
  module_name_from_path,
  load_module,
  load_entrypoint )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
def test_load_entrypoint():
  root = Path(__file__).parent / 'pkg_base'


  f = load_entrypoint('pkgaux:dist_prep', root)

  assert callable(f)

  with raises( ValueError ):
    load_entrypoint('pkgaux:not_an_attr', root)



import os
import os.path as osp
import tempfile
import shutil
import subprocess
from pathlib import Path

from ..validate import (
  validating,
  ValidationError,
  ValidPathError,
  FileOutsideRootError )

from ..load_module import EntryPoint

from ..path import (
  subdir )

#+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
class Builder:
  """Run build setup, compile, install commands

  Parameters
  ----------
  root : str | pathlib.Path
    Path to root project directory
  targets : :class:`pyproj_build <partis.pyproj.pptoml.pyproj_targets>`
  logger : logging.Logger
  """
  #-----------------------------------------------------------------------------
  def __init__(self,
    pyproj,
    root,
    targets,
    logger):

    self.pyproj = pyproj
    self.root = Path(root).resolve()
    self.targets = targets
    self.logger = logger
    self.target_paths = [
      dict(
        src_dir = target.src_dir,
        build_dir = target.build_dir,
        prefix = target.prefix )
      for target in targets ]

  #-----------------------------------------------------------------------------
  def __enter__(self):

    try:
      for i, (target, paths) in enumerate(zip(self.targets, self.target_paths)):
        if not target.enabled:
          self.logger.info(f"Skipping targets[{i}], disabled for environment markers")
          continue

        # check paths
        for k in ['src_dir', 'build_dir', 'prefix']:
          with validating(key = f"tool.pyproj.targets[{i}].{k}"):

            rel_path = paths[k]

            abs_path = (self.root / rel_path).resolve()

            if not subdir(self.root, abs_path, check = False):
              raise FileOutsideRootError(
                f"Must be within project root directory:"
                f"\n  file = \"{abs_path}\"\n  root = \"{self.root}\"")


            paths[k] = abs_path

        src_dir = paths['src_dir']
        build_dir = paths['build_dir']
        prefix = paths['prefix']

        with validating(key = f"tool.pyproj.targets[{i}].src_dir"):
          if not src_dir.exists():
            raise ValidPathError(f"Source directory not found: {src_dir}")

        with validating(key = f"tool.pyproj.targets[{i}]"):
          if subdir(build_dir, prefix, check = False):
            raise ValidPathError(f"'prefix' cannot be inside 'build_dir': {build_dir}")

        for k in ['build_dir', 'prefix']:
          with validating(key = f"tool.pyproj.targets[{i}].{k}"):
            dir = paths[k]

            if dir == self.root:
              raise ValidPathError(f"'{k}' cannot be root directory: {dir}")

            dir.mkdir( parents = True, exist_ok = True )

        entry_point = EntryPoint(
          pyproj = self,
          root = self.root,
          name = f"tool.pyproj.targets[{i}]",
          logger = self.logger,
          entry = target.entry )

        self.logger.info(f"Running build[{i}]")
        self.logger.info(f"Source dir: {src_dir}")
        self.logger.info(f"Build dir: {build_dir}")
        self.logger.info(f"Prefix: {prefix}")

        entry_point(
          options = target.options,
          src_dir = src_dir,
          build_dir = build_dir,
          prefix = prefix,
          setup_args = target.setup_args,
          compile_args = target.compile_args,
          install_args = target.install_args,
          build_clean = target.build_clean )

    except:
      self.build_clean()
      raise

  #-----------------------------------------------------------------------------
  def __exit__(self, type, value, traceback):
    self.build_clean()

    # do not handle any exceptions here
    return False

  #-----------------------------------------------------------------------------
  def build_clean(self):
    for i, (target, paths) in enumerate(zip(self.targets, self.target_paths)):
      build_dir = paths['build_dir']

      if build_dir is not None and build_dir.exists() and target.build_clean:
        self.logger.info(f"Removing build dir: {build_dir}")
        shutil.rmtree(build_dir)
